-- character encoding: ISO-8859-2
-- fileformat=UNIX

-- Nono-solver
-- Zadanie nr 2 na pracowni z Programowania 2008
-- Mateusz Styrczula
-- 11.05.2008

{-

    - Idea dziaania programiku (oraz przepyw sterowania):
      -----------------------------------------------------
        - Pierwsza wywoywana jest funkcja "main", ktra jest odpowiedzialna
            za wczytanie danych ze standardowego wejcia ("read_list")
            i umieszczenie ich w odpowiednich abstrakcyjnych strukturach danych
            (typ danych "Nonogram"). adne sprawdzanie poprawnoci danych
            nie jest wykonywane (zatem dla nieprawidowych danych
            dziaanie programu nie jest okrelone).
        - Funkcja "main" wywouje "nono_solver", ktry:
            - uruchamia algorytm dedukujcy ktre pola musz by
                koniecznie czarne ("nono_sure_black") zarwno
                dla wierszy, jak i dla kolumn,
            - takie wydedukowane, czciowo ukonkretnione nonogramy
                s klejone (funkcj "glue") w jeden czciowo
                ukonkretniony nonogram,
            - nastpnie wywoywany jest algorytm, ktry metod
                brute-force (z uwzgldnieniem informacji
                zawartych w czciowo ukonkretnionej planszy
                rozwizywanego nonogramu) szuka wszystkich moliwych
                rozwiza speniajcych specyfikacj. Jest to funkcja
                "nono_generator" ktra:
                - generuje kolejne wiersze (funkcj "rows_generator"),
                - taki wygenerowany wiersz dokada do generowanego rozwizania,
                - funkcj "nono_checker" sprawdza, czy dotychczasowe
                    rozwizanie ma szanse by poprawne.
        - majc wygenerowane nonogramy (opisane jako "Grid Square"),
            funkcja "main" wywouje "asciiartize" aby wywietli
            rozwizania,
        - ostatecznie wywietlana jest linia informujca o liczbie
            wygenerowanych rozwiza.


        Dooyem wszelkich stara, aby programik mia moduow budow.
    W szczeglnoci moliwe jest atwe dopisanie innych heurystyk
    ni tylko "koniecznie czarne" i wykonanie ich przed ostatecznym
    algorytmem brute-force'owym (po napisaniu takiej heurystyki
    wystarczy jej wywoanie doda w funkcji "nono_solver" przed wywoaniem
    funkcji "nono_generator" i - w razie gdyby jej dziaanie byo
    niezalene od dotychczas przeprowadzonych dedukcji - za pomoc
    funkcji "glue" doklei wynik jej dziaania do ju czciowo
    ukonkretnionego nonogramu).

 -}




import System.IO
import Control.Monad
import Data.List


-- zadanie jest opisane w nastpujcy sposb:
data Nonogram = Nonogram {
    -- szeroko nonogramu
    width :: Int,
    -- wysoko nonogramu
    height :: Int,
    -- listy czarnych obszarw w kolumnach
    width_list :: [[Int]],
    -- listy czarnych obszarw w wierszach
    height_list :: [[Int]]
}


-- podstawowy typ danych - kwadratowe pole, ktre moe by "Czarne",
-- "Biae", lub "jeszcze nieokrelone"
data Square = Black | White | Unknown deriving (Eq, Enum)
instance Show Square where
    show Black = "#"
    show White = " "
    show Unknown = "."


-- wiersz nonogramu
type Row a = [a]


-- nonogram skada si z wierszy - typem wynikowym bdzie taka "tablica komrek"
type Grid a = [Row a]


-- dla uatwienia zdefiniujmy sobie taki typ
-- (eby nie mczy si z myleniem o wielokrotnie zagniedanych listach)
type Generator a = [a]




-- implementacja heurystyki "koniecznie pene"
-- [ http://pl.wikipedia.org/wiki/Malowanie_liczbami#Koniecznie_pe.C5.82ne ]
-- funkcja dla zadanej specyfikacji nonogramu zwraca plansz z zaczernionymi
-- komrkami, ktre w kadym poprawnym przypadku musz by czarne
nono_sure_black :: Nonogram -> Grid Square
nono_sure_black (Nonogram width height width_list height_list) =
    nonosb height_list where


        -- rekurencyjnie dla kadego wiersza ustala ktre komrki
        -- powinny by "koniecznie czarne"
        nonosb [] = []
        nonosb (h:hs) =
            ((gnrh ++ replicate len 0)
            `insect`
            (replicate len 0 ++ gnrh)) : nonosb hs where

                -- gnr - opisana niej
                gnrh = gnr 1 h
                -- total - ile komrek zajmuj czarne bloki w wierszu
                -- (cznie z biaymi polami pomidzy nimi)
                total = sum h + (length h) - 1
                -- len - ilo komrek biaych "na kracach"
                len = width - total


        -- funkcja generuje dla danej specyfikacji wiersza [Int]
        -- wiersz w "postaci liczbowej".
        -- Przykad: 
        --         specyfikacja - [3,2,4]
        --      wynik dziaania - [1,1,1,0,2,2,0,3,3,3,3] 
        -- Funkcja jako separator blokw (odpowiednik White'a) wstawia "0".
        -- Bloki s numerowane poczwszy od wartoci argumentu "act".
        gnr :: Int -> [Int] -> Row Int
        gnr _ [] = []
        gnr act (bs:bss) =
            replicate bs act ++
            if null bss
                then gnr (act+1) bss
                else 0 : gnr (act+1) bss


        -- dla dwch list rwnej dugoci w "postaci liczbowej"
        -- zwraca ich "przecicie", tzn. jeli na odpowiednich pozycjach
        -- znajduje si czarne pole (liczba > 0), to w "przeciciu" te bdzie
        -- czarne pole, w p.p. "pole nieznane" (za wikipedi: "...Wanym jest
        -- aby bra pod uwag tylko te pola, ktre pokryte s tym samym blokiem...")
        insect :: Row Int -> Row Int -> Row Square
        (l:ls) `insect` (r:rs)
            | l == r && l /= 0 = Black : (ls `insect` rs)
            | otherwise        = Unknown : (ls `insect` rs)
        _ `insect` _    = []




-- Funkcja sprawdzajca dla danej specyfikacji Nonogramu
-- czy podany nonogram (jako "tablica komrek" - Grid Square) spenia
-- t specyfikacj dla kolumn (zaoenie jest takie, e podany nonogram
-- spenia specyfikacj dla wierszy).
nono_checker :: Nonogram -> Grid Square -> Bool
nono_checker (Nonogram _ _ width_list _) grid =
    -- sprawdzenie wykonaj dla transponowanego nonogramu
    -- (chcemy uzyska list kolumn z listy wierszy)
    nonche width_list (transpose grid) where


        -- Funkcja "nonche" bierze list list czarnych obszarw w kolumnach
        -- oraz list kolumn i sprawdza rekurencyjnie, czy kada z tych kolumn spenia
        --- odpowiadajc jej list czarnych obszarw.
        nonche [] _ = True
        nonche (w:ws) (c:cs)
            | c `fulfills` w = nonche ws cs
            | otherwise      = False


        -- Funkcja "fulfills" sprawdza, czy kolumna (bdca lewym jej argumentem)
        -- potencjalnie spenia odpowiadajc jej specyfikacj - list czarnych pl
        -- (bdc jej prawym argumentem). "Potencjalnie spenia" oznacza,
        -- e sprawdzana kolumna moe by "niepena", np. kolumna [#,#,.,.,#]
        -- "potencjalnie spenia" specyfikacj [2,3].
        [] `fulfills` []          = True
        c@(_:_) `fulfills` []     = Black `notElem` c
        [] `fulfills` (_:_)       = False
        c@(_:_) `fulfills` (b:bs) =
            length block == b  &&  rest `fulfills` bs
            ||
            length block <= b  &&  null rest where
                without_leading_whites = dropWhile (/= Black) c
                (block, rest) = span (== Black) without_leading_whites




-- generator wierszy, bierze:
--      szerorko wiersza - width,
--      list dugoci kolejnych blokw w tym wierszu (specyfikacj),
--      czciowo ukonkretniony wiersz,
-- Wynikiem dziaania tej funkcji s wszystkie wiersze podanej dugoci speniajce
-- zadan specyfikacj (list dugoci kolejnych blokw w tym wierszu).
rows_generator :: Int -> [Int] -> Row Square -> Generator (Row Square)
rows_generator width _ [] = []
rows_generator width [] partial = return $ replicate width White
rows_generator width block_sizes@(bs:bss) partial@(sq:sqs)
    -- jeli blok jest wikszy ni dugo wiersza,
    -- to oczywicie nie mona go skonstruowa
    | width < bs    = []
    -- poniej rozpatrzone s 3 gwne przypadki:
    -- (1) - czciowo ukonkretniony wiersz zaczyna si czarn komrk,
    -- wic konstruowany blok czarnych komrek musi si z ni pokrywa
    -- (zaczyna si od niej)
    | sq == Black   = do
        -- naley rozrni sytuacj, kiedy s jeszcze jakie bloki do konstruowania,
        -- a kiedy nie (w pierwszym przypadku mona ju pozosta przestrze do koca
        -- wiersza wypeni "White'ami", w p.p. naley rekurencyjnie skonstruowa
        -- pozostae bloki)
        if null bss  
            then return $ block_of_blacks ++ replicate (width-bs) White
            else do
                row <- rows_generator (width-bs-1) bss (drop (bs+1) partial)
                -- oczywicie pomidzy skonstruowanymi rekurencyjnie blokami
                -- a blokiem skonstruowanym w biecym wywoaniu naley wstawi
                -- jednego "White'a"
                return $ block_of_blacks ++ (White : row)
    -- (2) - czciowo ukonkretniony wiersz zaczyna si bia komrk, wic aden
    -- blok czarnych komrek nie moe si z ni pokrywa
    | sq == White   = do
        row <- rows_generator (width-1) block_sizes sqs
        return $ (White : row)
    -- (3) - czciowo ukonkretniony wiersz nie daje adnej wskazwki w kwestii
    -- aktualnej komrki, zatem trzeba rozway dwa przypadki...
    | otherwise     = do
        -- ... gdy pierwsza komrka jest biaa lub ...
        row <- rows_generator (width-1) block_sizes sqs
        return $ (White : row)
        -- ... gdy pierwsza komrka jest czarna
        `mplus`
        if null bss
            -- skoro konstruujemy ostatni blok (bo bss jest pusty) to w
            -- czciowo ukonkretnionym wierszu za konstruowanym wanie blokiem
            -- nie moe wystpi ju adna czarna komrka...
            then if Black `notElem` drop bs partial
                then return $ block_of_blacks ++ replicate (width-bs) White
                -- ...ale gdyby jednak tak byo, to znaczy, e bieca komrka musi
                -- by biaa, ale taki przypadek ju zosta obsuony
                else []  
            -- jeli nie konstruujemy ostatniego bloku (bss nie jest pusty), to
            -- naley si upewni, czy nastpna komrka wystpujca po wanie
            -- konstruowanym bloku nie jest czarna...
            else if drop bs partial /= []  &&  Black /= (head $ drop bs partial)
                then do
                    row <- rows_generator (width-bs-1) bss (drop (bs+1) partial)
                    return $ block_of_blacks ++ (White : row)
                -- ...bo, gdyby tak byo, to bieca komrka musi by biaa a to
                -- ju byo obsuone
                else [] where  

                
        block_of_blacks = replicate bs Black




-- Generuje wszystkie kolejne nonogramy speniajce
-- "specyfikacj wierszow" (height_list) oraz "specyfikacj kolumnow" (width_list).
-- Oprcz opisu zadania (Nonogram) bierze take (by moe czciowo ukonkretnion)
-- plansz reprezentujc rozwizywany nonogram.
nono_generator :: Nonogram -> Grid Square -> Generator (Grid Square)
nono_generator (Nonogram width height width_list height_list) partial_grid
    -- nonogram bdzie budowany "od ostatniego wiersza"
    -- (jest to spowodowane kolejnoci dokadania kolejnych
    -- wierszy do czciowo zbudowanego nonogramu - poniej,
    -- miejsce oznaczone: <*>)
    = nongen [] (reverse height_list) (reverse partial_grid) where


        -- "nongen" generujc kolejny wiersz nonogramu sprawdza
        -- "kolumnow poprawno" nonogramu do tej pory wygenerowanego.
        -- Sprawdzanie takiej "czciowej poprawnoci" odcina bardzo duo
        -- niepotrzebnych prb ("lepych zaukw") rozwizania nonogramu.
        -- Argumentami s odpowiednio:
        --     grid   - konstruowany nonogram
        --     (h:hs) - odwrcona lista specyfikacji wierszowych
        --     (r:rs) - czciowo ukonkretniony (uytymi wczeniej heurystykami)
        --              nonogram
        nongen :: Grid Square -> [[Int]] -> Grid Square -> Generator (Grid Square)
        nongen grid [] _ = return grid
        nongen grid (h:hs) (r:rs) = do
            -- wygeneruj wiersz (o dugoci "width", speniajcy specyfikacj "h"
            -- i pokrywajcy si z czciowo wypenionym wierszem "r")
            row <- rows_generator width h r
            let partial_nono = (row:grid)  -- <*>
            -- upewnij si, e dotychczas wygenerowany kawaek nonogramu
            -- spenia "specyfikacj kolumnow"
            guard $ nono_checker (Nonogram width height rev_width_list height_list)
                                 (reverse partial_nono)
            -- rekurencyjnie zbuduj pozostae wiersze nonogramu
            nongen partial_nono hs rs


        rev_width_list = map reverse width_list




-- "Serce" programu - funkcja pobierajca opis zadania i generujca rozwizania
-- w postaci typu wynikowego opisanego wyej.
nono_solver :: Nonogram -> Generator (Grid Square)
nono_solver (Nonogram width height width_list height_list) = do
    -- Pierwszym krokiem jest zastosowanie heurystyki "koniecznie czarne" dla wierszy...
    let grid = nono_sure_black (Nonogram width height width_list height_list) 
               -- ... oraz ...
               `glue`
               -- ... dla kolumn ("odwracamy" nonogram, a potem transponujemy wynik).
               (transpose $ nono_sure_black (Nonogram height width height_list width_list))
    -- Dla tak wstpnie wypenionej planszy mona uruchomi funkcj przeszukujc
    -- wszystkie moliwe (prawidowe i speniajce to wstpne wypenienie) rozwizania.
    nono_generator (Nonogram width height width_list height_list) grid where


        -- Funkcja "sklejajca" dwie wstpnie wypenione plansze.
        -- Rekurencyjnie przechodzi po odpowiadajcych sobie wierszach
        -- i skleja je. Przykad:
        --     [[#,#,.,., ,.]] `glue` [[.,#,.,#,., ]] = [[#,#,.,#, , ]]
        glue :: Grid Square -> Grid Square -> Grid Square
        [] `glue` [] = []
        (l:ls) `glue` (r:rs) = (
                zipWith (\x y ->
                    if x == Black || y == Black
                        then Black
                        else if x == White || y == White
                            then White
                            else Unknown
                ) l r
            ) : (ls `glue` rs)
        _ `glue` _ = []




-- Punkt "wejciowy" caego programu:
-- Funkcja jest odpowiedzialna za wczytanie danych (ze standardowego wejcia), 
-- wywoanie funkcji rozwizujcej zagadk oraz wypisanie wyniku
-- (ascii-art'owego obrazka na standardowe wyjcie).
-- Poprawno danych wejciowych nie jest w aden sposb weryfikowana.
main :: IO ()
main = do
    -- wczytanie szerokoci i wysokoci nonogramu
    (nono_width, nono_height) <- readLn
    -- wczytanie listy czarnych obszarw w kolumnach
    nono_width_list <- read_list
    -- wczytanie listy czarnych obszarw w wierszach
    nono_height_list <- read_list
    -- rozwizanie amigwki
    let grids = nono_solver (Nonogram nono_width nono_height nono_width_list nono_height_list)
    -- "pretty-printing" kadego rozwizania zgodnie ze specyfikacj zadania
    mapM_ asciiartize grids
    -- i na koniec informacja o tym ile si udao wygenerowa rozwiza
    putStrLn $ show (length grids) ++ " solution(s)." where


    -- Funkcja czytajca ze standardowego wejcia linie
    -- i interpretujca kad z nich jako [Int] (list Intw).
    -- Koczy swoje dziaanie albo gdy natrafi na koniec pliku (EOF),
    -- albo gdy zostanie przeczytany pusty wiersz.
    read_list = do
        eof <- isEOF
        -- jeli nastpi koniec pliku to...
        if eof
            -- ...zakocz dziaanie, a...
            then return []
            -- ... w przeciwnym przypadku ...
            else do
                -- ... wczytaj lini
                l <- getLine
                -- jeli linia jest pusta, ...
                if l == ""
                    --  ... to zakocz dziaanie
                    then return []
                    -- w p.p. ...
                    else do
                        -- wczytaj pozostae linie i...
                        ls <- read_list
                        -- ... zwr list linii interpretowanych jako listy intw
                        return ((read l :: [Int]) : ls)


    -- "pretty-printer" - rekurencyjnie wypisuje (na standardowe wyjcie) wiersze
    asciiartize [] = putStrLn "--------------------------------"
    asciiartize (row:rows) = do
        -- Kada komrka w licie reprezentujcej linie ("Black", "White" lub "Unknown")
        -- zamieniana jest na odpowiedni symbol typu Char ("head . show"), a nastpnie
        -- taka lista [Char] (bdca zatem String'iem) wypisywana jest na standardowe wyjcie.
        putStrLn $ map (head . show) row
        asciiartize rows

